/*****************************************************************************
 *
 * FILE:	portscan.c
 * DESCRIPTION:	TCP/UDP Port Scanner (Another Implementation)
 * DATE:	Fri, Jun 16 2006
 * UPDATED:	Thu, May  1 2008
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MagickWorX.COM
 * URL:		http://www.MagickWorX.COM/
 * COPYRIGHT:	(c) 2006-2008 졿Kouichi ABE (WALL), All rights reserved.
 * NOTES:	This code dirives from "Beholder: Easy TCP/UDP Port Scanner"
 * LICENSE:
 *
 *  Copyright (c) 2006-2008 Kouichi ABE (WALL) <kouichi@MagickWorX.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 * $Id: portscan.c,v 1.1 2008/05/01 13:31:26 kouichi Exp $
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/if_ether.h>
#include <arpa/inet.h>
#include <net/ethernet.h>
#include <net/if.h>
#include <signal.h>
#include <errno.h>
#include "netfd.h"

/*****************************************************************************
 *
 *	Program and Copyright
 *
 *****************************************************************************/
#define	SOFTWARE_NAME	"portscan"
static const char	Program[]   = SOFTWARE_NAME;
static const char	Copyright[] =
	"Copyright (c) 2006-2008 Kouichi ABE (WALL), All rights reserved.";

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
/*
 * Numeric release version identifier:
 * MNNFFPPS: major minor fix patch status
 * The status nibble has one of the values 0 for development,
 * 1 to e for betas 1 to 14, and f for release.
 * The patch level is exactly that.
 */
#define PORTSCAN_VERSION_NUMBER	0x10100000L
#define PORTSCAN_VERSION	SOFTWARE_NAME "/1.1"
#define PORTSCAN_VERSION_TEXT	PORTSCAN_VERSION " (2008/05/01)"
#define	PORTSCAN_URI		"http://www.MagickWorX.COM/products/beholder/"

/*****************************************************************************/

#define	PORTSCAN_PORT_MIN	(1)
#define	PORTSCAN_PORT_MAX	(65535)
#define	PORTSCAN_TIMEOUT	(500)	/* msec */
/* XXX: UDP ξ 1000[msec] ֤Ȥɤ */
#define	PORTSCAN_TIMEOUT_UDP	(1000)	/* msec */
#define	PORTSCAN_RETRY_MAX	(1)	/* UDP åκƻ */

#define	DEF_PORT_MIN		PORTSCAN_PORT_MIN
#define	DEF_PORT_MAX		PORTSCAN_PORT_MAX

#define	DEF_NETIF_DEVICE	"em0"

#define	SIZE_ETHERHDR	(sizeof(struct ether_header))
#define	SIZE_IPHDR	(sizeof(struct ip))
#define	SIZE_ICMPHDR	(sizeof(struct icmp))
#define	SIZE_TCPHDR	(sizeof(struct tcphdr))
#define	SIZE_UDPHDR	(sizeof(struct udphdr))

struct timer {
  struct timeval	start;
  struct timeval	end;
  struct timeval	expend;
};

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void	version(void);
static void	usage(void);
static void	cleanup(int);
static void	set_sigtrap(void);

static void	printer(const char * ipaddr, const uint16_t port,
			const uint8_t protocol, const char * state);
static int	get_my_ipaddr(const char * netif, in_addr_t * my_ipaddr);
static int	checker(uint8_t *, uint16_t, const char *, uint16_t,
			const char *, uint16_t, uint8_t);
static int	shoot(netfd_t *, netfd_t *, const char *, uint16_t,
		      const char *, uint16_t, bool, netfd_time_t);
static int	portscan(uint32_t ip_lo, uint32_t ip_hi, uint16_t port_lo,
			 uint16_t port_hi, bool udp_proto, const char * netif,
			 netfd_time_t timeout);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static struct timer	timer;
static bool		show_all = false;

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/
static void
version(void)
{
  fprintf(stderr, "%s\n%s\n\n", PORTSCAN_VERSION_TEXT, Copyright);
  exit(64);
}

static void
usage(void)
{
  fprintf(stderr, "%s\n%s\n%s\n\n", PORTSCAN_VERSION_TEXT, Copyright, PORTSCAN_URI);
  fprintf(stderr,
	  "usage: %s [-i <device>][-t <timeout>][-a][-u] <ipaddr> [<port>]\n\n",
	  Program);
  fprintf(stderr, "\n    [ examples ]\n");
  fprintf(stderr, "       %s 192.168.1.1\n", Program);
  fprintf(stderr, "       %s 192.168.1.1 80\n", Program);
  fprintf(stderr, "       %s 192.168.1.1 1-1024\n", Program);
  fprintf(stderr, "       %s 172.16.1.100-172.16.1.200\n", Program);
  fprintf(stderr, "       %s 192.168.1.0/24\n", Program);
  fprintf(stderr, "\n");
  exit(64);
}

/*****************************************************************************/

static void
cleanup(signo)
	int	signo;
{
  exit(signo);
}

static void
set_sigtrap(void)
{
  struct sigaction	act;

  sigemptyset(&act.sa_mask);
  act.sa_flags	 = 0;
  act.sa_handler = (void *)cleanup;
  sigaction(SIGINT,  &act, NULL);
  sigaction(SIGTERM, &act, NULL);
  sigaction(SIGQUIT, &act, NULL);
  sigaction(SIGSEGV, &act, NULL);
  sigaction(SIGBUS,  &act, NULL);
  sigaction(SIGILL,  &act, NULL);
}

/*****************************************************************************/

static void
printer(ipaddr, port, protocol, state)
	const char *	ipaddr;		/* IP address */
	const uint16_t	port;		/* open port number */
	const uint8_t	protocol;	/* "tcp" | "udp" */
	const char *	state;		/* "open" | "timedout" */
{
  struct protoent *	p;

  p = getprotobynumber(protocol);
  if (p != NULL) {
    fprintf(stdout, "%s %5hu/%s\t[%s]\n", ipaddr, port, p->p_name, state);
  }
  else {
    fprintf(stdout, "%s %5hu/%u\t[%s]\n", ipaddr, port, protocol, state);
  }
}

/*****************************************************************************/

static int
get_my_ipaddr(netif, my_ipaddr)
	const char *	netif;
	in_addr_t *	my_ipaddr;	/* returned */
{
  int		sd;
  struct ifreq	ifr;
  int		status = -1;

  (void)strcpy(ifr.ifr_name, netif);
  sd = socket(AF_INET, SOCK_DGRAM, 0);
  if (sd != -1) {
    if (ioctl(sd, SIOCGIFADDR, &ifr) == 0) {
      *my_ipaddr = ((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr.s_addr;
      status	 = 0;
    }
    close(sd);
  }

  return status;
}

/*****************************************************************************/

/*
 * -1: not open
 *  0: open
 *  1: another packets
 */
static int
checker(pktbdy, caplen, chk_saddr, chk_sport, chk_daddr, chk_dport, protocol)
	uint8_t *	pktbdy;
	uint16_t	caplen;
	const char *	chk_saddr;
	uint16_t	chk_sport;
	const char *	chk_daddr;
	uint16_t	chk_dport;
	uint8_t		protocol;
{
  struct ether_header *	ether;
  int			status = 1;

  ether = (struct ether_header *)pktbdy;
  if (ntohs(ether->ether_type) == ETHERTYPE_IP) {
    struct ip *	ip;
#if	0
    size_t	offset = sizeof(struct ether_header);
#else
    size_t	offset = ETHER_HDR_LEN;
#endif
    uint8_t	src_addr[NETFD_IP_ADDR_SIZE];
    uint8_t	dst_addr[NETFD_IP_ADDR_SIZE];

    ip = (struct ip *)(pktbdy + offset);
    memset(src_addr, 0, sizeof(src_addr));
    memset(dst_addr, 0, sizeof(dst_addr));
    strlcpy(src_addr, inet_ntoa(ip->ip_src), sizeof(src_addr));
    strlcpy(dst_addr, inet_ntoa(ip->ip_dst), sizeof(dst_addr));

    offset += (ip->ip_hl << 2);
    switch (ip->ip_p) {
      case IPPROTO_ICMP: {
	  struct icmp *		icmp;
	  struct ip *		ih;	/* internet header */
	  struct udphdr *	udp;	/* 64 bits of original data datagram */

	  status = -1;
	  if (caplen - offset < ICMP_MINLEN) {
	    fprintf(stderr, "%s sent truncated icmp: %d",
		    inet_ntoa(ip->ip_src), caplen - offset);
	    break;
	  }
	  icmp = (struct icmp *)(pktbdy + offset);
#if	DEBUG
	  fprintf(stdout, "ICMP: ID: %hu Length: %hu TTL: %d\n"
		  "\t%s -> %s (%d.%d)\n",
		  ntohs(ip->ip_id), ntohs(ip->ip_len), ip->ip_ttl,
		  src_addr, dst_addr, icmp->icmp_type, icmp->icmp_code);
#endif	/* DEBUG */
	  offset += ICMP_MINLEN;
	  switch (icmp->icmp_type) {
	    case ICMP_UNREACH:
	    case ICMP_TIMXCEED:
	    case ICMP_PARAMPROB:
	    case ICMP_SOURCEQUENCH:
	    case ICMP_REDIRECT:
	      ih  = (struct ip *)(pktbdy + offset);
	      udp = (struct udphdr *)(pktbdy + offset + SIZE_IPHDR);
	      if (icmp->icmp_type == ICMP_TIMXCEED) {
		if (show_all) {
		  printer(chk_daddr, chk_dport, protocol, "no route");
		}
	      }
	      else if (icmp->icmp_type == ICMP_UNREACH) {
		if (icmp->icmp_code == ICMP_UNREACH_PORT) {
		  if (show_all) {
		    if (ntohs(udp->uh_sport) == chk_sport &&
			ntohs(udp->uh_dport) == chk_dport) {
		      printer(chk_daddr, chk_dport, protocol, "closed");
		    }
		  }
		}
		else if (icmp->icmp_code == ICMP_UNREACH_FILTER_PROHIB) {
		  printer(chk_daddr, chk_dport, protocol, "filtered");
		}
	      }
#if	DEBUG
	      if (protocol == IPPROTO_UDP) {
		fprintf(stderr, "DEBUG[icmp] %hu -> %hu\n",
			ntohs(udp->uh_sport), ntohs(udp->uh_dport));
	      }
#endif	/* DEBUG */
	      break;
	    default:
	      break;
	  }
	  return status;
	}
	break;
      case IPPROTO_TCP:
	if (strcmp(chk_saddr, dst_addr) == 0 &&
	    strcmp(chk_daddr, src_addr) == 0) {
	  struct tcphdr *	tcp;

	  if (caplen - offset < SIZE_TCPHDR) {
	    fprintf(stderr, "%s sent truncated tcp: %d",
		    inet_ntoa(ip->ip_src), caplen - offset);
	    status = -1;
	    break;
	  }
#define	TH_SYNACK	(TH_SYN|TH_ACK)
#define	TH_RSTACK	(TH_RST|TH_ACK)
	  tcp = (struct tcphdr *)(pktbdy + offset);
	  if (ntohs(tcp->th_sport) == chk_dport) {
	    switch (tcp->th_flags) {
	      case TH_SYNACK:
		printer(chk_daddr, chk_dport, protocol, "open");
		status = 0;
		break;
	      case TH_RSTACK:
		if (show_all) {
		  printer(chk_daddr, chk_dport, protocol, "closed");
		}
		status = -1;
		break;
	      default: break;
	    }
	  }
#if	DEBUG
	  do {
	    char	s[9];

	    memset(s, '*', sizeof(s));
	    if (tcp->th_flags & TH_FLAGS) {
	      if (tcp->th_flags & TH_FIN)  { s[0] = 'F'; }
	      if (tcp->th_flags & TH_SYN)  { s[1] = 'S'; }
	      if (tcp->th_flags & TH_RST)  { s[2] = 'R'; }
	      if (tcp->th_flags & TH_PUSH) { s[3] = 'P'; }
	      if (tcp->th_flags & TH_ACK)  { s[4] = 'A'; }
	      if (tcp->th_flags & TH_URG)  { s[5] = 'U'; }
	      if (tcp->th_flags & TH_ECE)  { s[6] = 'E'; }
	      if (tcp->th_flags & TH_CWR)  { s[7] = 'C'; }
	    }
	    s[sizeof(s) - 1] = '\0';

	    fprintf(stdout, "TCP: ID: %hu Length: %hu TTL: %d "
		    "Seq#: %u ACK#: %u [%s]\n"
		    "\t%s:%hu -> %s:%hu\n",
		    ntohs(ip->ip_id), ntohs(ip->ip_len), ip->ip_ttl,
		    ntohl(tcp->th_seq), ntohl(tcp->th_ack), s,
		    src_addr, ntohs(tcp->th_sport),
		    dst_addr, ntohs(tcp->th_dport));
	  } while (0);
#endif	/* DEBUG */
	  return status;
	}
	break;
      case IPPROTO_UDP:
#if	DEBUG
	if (strcmp(chk_saddr, dst_addr) == 0 &&
	    strcmp(chk_daddr, src_addr) == 0) {
	  struct udphdr *	udp;

	  if (caplen - offset < SIZE_UDPHDR) {
	    fprintf(stderr, "%s sent truncated udp: %d",
		    inet_ntoa(ip->ip_src), caplen - offset);
	    status = -1;
	    break;
	  }
	  udp = (struct udphdr *)(pktbdy + offset);
	  fprintf(stdout, "UDP: ID: %hu Length: %hu TTL: %d\n"
		  "\t%s:%hu -> %s:%hu\n",
		  ntohs(ip->ip_id), ntohs(ip->ip_len), ip->ip_ttl,
		  src_addr, ntohs(udp->uh_sport),
		  dst_addr, ntohs(udp->uh_dport));
	}
#endif	/* DEBUG */
	break;
      default:
	fprintf(stderr, "unexpected protocol: %d\n", ip->ip_p);
	status = -1;
	break;
    }
#if	0
    if (strcmp(chk_saddr, dst_addr) == 0 && strcmp(chk_daddr, src_addr) == 0) {
      status = 0;
    }
#endif
  }

  return status;
}

static int
shoot(wfd, rfd, src_addr, src_port, dst_addr, dst_port, udp_proto, timeout)
	netfd_t *	wfd;
	netfd_t *	rfd;
	const char *	src_addr;
	uint16_t	src_port;
	const char *	dst_addr;
	uint16_t	dst_port;
	bool		udp_proto;
	netfd_time_t	timeout;
{
  int			status	= 1;
  ssize_t		ip_size	= 0;
  uint8_t *		ip_pkt;
  int			rv;
  struct addrinfo *	ai;

  if ((rv = netfd_host_serv(dst_addr, NULL, 0, 0, 0, &ai)) != 0) {
    fprintf(stderr, "%s\n", gai_strerror(rv));
    return -1;
  }

#if	DEBUG
  fprintf(stderr, "DEBUG[shoot] %s:%hu -> %s:%hu\n",
	  src_addr, src_port, dst_addr, dst_port);
#endif	/* DEBUG */

#define	MY_TTL	64
  if (udp_proto) {
    ip_size = netfd_build_udp_packet(src_addr, src_port, dst_addr, dst_port,
				     MY_TTL, 0, &ip_pkt);
  }
  else {
    ip_size = netfd_build_tcp_packet(src_addr, src_port, dst_addr, dst_port,
				     TH_SYN, 0, 0, 4096, MY_TTL, 0, &ip_pkt);
  }

  if (ip_size > 0) {
    unsigned int	c = 0;	/*  retry counter */

    do {	/* XXX: this block provides special regulation to IP header */
      struct ip *	ip = (struct ip *)ip_pkt;

      ip->ip_len = ntohs(ip->ip_len);
      ip->ip_off = ntohs(ip->ip_off);
    } while (0);

again:
    if (netfd_sendto(wfd, ip_pkt, ip_size, 0, ai->ai_addr, ai->ai_addrlen,
		     NETFD_NO_TIMEOUT) == ip_size) {
      while (status > 0) {
	ssize_t	size;
	uint8_t	buf[BUFSIZ*64];

	size = netfd_read(rfd, buf, sizeof(buf), timeout);
	if (size > 0) {
	  uint8_t *		p;
	  struct bpf_hdr *	bhp;

	  for (p = buf; p < buf + size;
	      p += BPF_WORDALIGN(bhp->bh_hdrlen + bhp->bh_caplen)) {
	    uint8_t *	pktbdy;

	    bhp    = (struct bpf_hdr *)p;
	    pktbdy = (uint8_t *)(p + bhp->bh_hdrlen);
	    status = checker(pktbdy, bhp->bh_caplen,
			     src_addr, src_port, dst_addr, dst_port,
			     udp_proto ? IPPROTO_UDP : IPPROTO_TCP);
	  }
	}
	else if (errno == ETIMEDOUT) {
	  if (udp_proto) {
	    if (c++ < PORTSCAN_RETRY_MAX) {
	      goto again;
	    }
	    printer(dst_addr, dst_port, IPPROTO_UDP, "open");
	    status = 0;
	  }
	  else {
	    printer(dst_addr, dst_port, IPPROTO_TCP, "timedout");
	    status = -1;
	  }
	}
	else {
	  fprintf(stderr, "DEBUG: size=%d errno=%d\n", size, errno);
	}
      }
      /*
       * XXX: TCP ξϤ RST ѥåȤ٤
       */
    }
    else {
      fprintf(stderr, "errno=%d %s\n", errno, strerror(errno));
      status = -1;
    }
    free(ip_pkt);
  }

  freeaddrinfo(ai);

  return status;
}

static int
portscan(uint32_t ip_lo, uint32_t ip_hi, uint16_t port_lo, uint16_t port_hi,
	 bool udp_proto, const char * netif, netfd_time_t timeout)
{
  uint16_t	flags = NETFD_BPF_IMMEDIATE|NETFD_BPF_HDRCMPLT|NETFD_BPF_NOSEESENT;
  netfd_t *	wfd;
  netfd_t *	rfd;
  int		n = 0;

  wfd = netfd_open_raw_ip(true);
  rfd = netfd_open_bpf(netif, flags, 0, false, 0, NULL);
  if (wfd != NULL && rfd != NULL) {
    in_addr_t	my_addr;

    if (get_my_ipaddr(netif, &my_addr) == 0) {
      struct in_addr	in;
      char		src_addr[sizeof("xxx:xxx:xxx:xxx")];
      uint16_t		src_port = 0;

      in.s_addr	= my_addr;
      inet_ntop(AF_INET, &in, src_addr, sizeof(src_addr));

      gettimeofday(&timer.start, (struct timezone *)NULL);
      do {
	char		dst_addr[sizeof("xxx:xxx:xxx:xxx")];
	uint32_t	dst_port;

	in.s_addr = ip_lo;
	inet_ntop(AF_INET, &in, dst_addr, sizeof(dst_addr));

	dst_port = port_lo;
	do {
	  int	status;

	  status = shoot(wfd, rfd,
			 src_addr, src_port, dst_addr, (uint16_t)dst_port,
			 udp_proto, timeout);
	  if (status == 0) {
	    n++;
	  }
	  dst_port = dst_port + 1;
	} while (dst_port <= (uint32_t)port_hi);
	ip_lo = htonl(ntohl(ip_lo) + 1);
      } while (ntohl(ip_lo) <= ntohl(ip_hi));
      gettimeofday(&timer.end, (struct timezone *)NULL);
      timersub(&timer.end, &timer.start, &timer.expend);
    }
    netfd_close(wfd);
    netfd_close(rfd);
  }

  return n;
}

/*****************************************************************************/

#define	ERRMSG_IP	"IP"
#define	ERRMSG_SUBNET	"subnet"

static int	get_ip_range(const char * addr,
			     uint32_t * ip_lo, uint32_t * ip_hi);
static int	check_ipaddr(const char * s);
static uint32_t	inet_addr_wrap(const char *, const char *);
static int	get_port_range(const char * port,
			       uint16_t * port_lo, uint16_t * port_hi);

static int
get_ip_range(addr, ip_lo, ip_hi)
	const char *	addr;
	uint32_t *	ip_lo;
	uint32_t *	ip_hi;
{
  register char *	cp;

  if ((cp = strchr(addr, '-')) != NULL && check_ipaddr(cp + 1) == 0) {
    *cp++  = '\0';
    *ip_lo = inet_addr_wrap(addr, ERRMSG_IP);
    *ip_hi = inet_addr_wrap(cp, ERRMSG_IP);
    if (ntohl(*ip_lo) > ntohl(*ip_hi)) {
      fprintf(stderr, "Invalid address range: %s-%s\n", addr, cp);
      exit(1);
    }
  }
  else if ((cp = strchr(addr, '/')) != NULL) {
    uint32_t	subnet;
    uint32_t	netmask;
    int		bitmask;

    *cp++   = '\0';
    subnet  = inet_addr_wrap(addr, ERRMSG_SUBNET);
    subnet  = ntohl(subnet);
    bitmask = (int)strtol(cp, (char **)NULL, 10);
    if (bitmask < 1 || bitmask > 32) {
      fprintf(stderr, "Invalid netmask: %s\n", cp);
      exit(1);
    }
    netmask = 0xffffffff << (32 - bitmask);
    *ip_lo  = htonl(subnet & netmask);
    *ip_hi  = htonl(subnet | ~netmask);
  }
  else {
    struct hostent *	hp;

    hp = gethostbyname2(addr, AF_INET);
    if (hp != NULL) {
      memcpy(ip_lo, hp->h_addr, (size_t)hp->h_length);
      *ip_hi = *ip_lo;
    }
    else {
      fprintf(stderr, "Invalid IP address: %s\n", addr);
      exit(1);
    }
  }

  return 0;
}

static int
check_ipaddr(s)
	const char *	s;
{
  const char	ipstr[] = "0123456789.";
  size_t	len;
  size_t	pos;

  len = strlen(s);
  pos = strspn(s, ipstr);
  return (len == pos ? 0 : -1);
}

static uint32_t
inet_addr_wrap(addr, mesg)
	const char *	addr;
	const char *	mesg;
{
  uint32_t	ip;

  ip = (uint32_t)inet_addr(addr);
  if (ip == INADDR_NONE) {
    fprintf(stderr, "Invalid %s address: %s\n", mesg, addr);
    exit(1);
  }

  return ip;
}

static int
get_port_range(port, port_lo, port_hi)
	const char *	port;
	uint16_t *	port_lo;
	uint16_t *	port_hi;
{
  register char *	cp;

  if ((cp = strchr(port, '-')) != NULL) {
    uint32_t	xmin;
    uint32_t	xmax;

    *cp++ = '\0';
    xmin = (uint32_t)strtol(port, (char **)NULL, 10);
    xmax = (uint32_t)strtol(cp,   (char **)NULL, 10);
    if (xmin > xmax || xmin < PORTSCAN_PORT_MIN || xmax > PORTSCAN_PORT_MAX) {
	fprintf(stderr, "Invalid port range: %s-%s\n", port, cp);
	exit(1);
    }
    *port_lo = (uint16_t)xmin;
    *port_hi = (uint16_t)xmax;
  }
  else {
    *port_lo = (uint16_t)strtol(port, (char **)NULL, 10);
    *port_hi = *port_lo;
  }

  return 0;
}

/*****************************************************************************/

int
main(argc, argv)
	int	argc;
	char *	argv[];
{
  register int	ch;
  int		opt_u	= false;
  const char *	opt_i	= DEF_NETIF_DEVICE;
  netfd_time_t	timeout	= PORTSCAN_TIMEOUT;
  uint32_t	ip_lo;	/* low IP address */
  uint32_t	ip_hi;	/* high IP address */
  uint16_t	port_lo	= DEF_PORT_MIN;	/* low port number */
  uint16_t	port_hi	= DEF_PORT_MAX;	/* high port number */
  const char *	addrs	= NULL;
  const char *	ports	= NULL;
  int		status	= -1;

  while ((ch = getopt(argc, argv, "aui:t:hv?")) != -1) {
    switch (ch) {
      case 'a': show_all = true; break;
      case 'u': opt_u = true; break;
      case 'i': opt_i = optarg; break;
      case 't': timeout = (netfd_time_t)strtol(optarg, (char **)NULL, 10); break;
      case 'v': version();
      default:	usage();
    }
  }
  argc -= optind;
  argv += optind;
  switch (argc) {
    case 2: ports = argv[1];
    case 1: addrs = argv[0]; break;
    default: usage();
  }

  if (netfd_init(NETFD_MODEL_TINY) == -1) {
    perror("netfd_init");
    exit(1);
  }

  get_ip_range(addrs, &ip_lo, &ip_hi);
  if (ports != NULL) {
    get_port_range(ports, &port_lo, &port_hi);
  }

  if (opt_u && timeout == PORTSCAN_TIMEOUT) {
    timeout = PORTSCAN_TIMEOUT_UDP;
  }

  set_sigtrap();
  status = portscan(ip_lo, ip_hi, port_lo, port_hi, opt_u, opt_i, timeout);

#if	DEBUG
  fprintf(stderr, "total: %d: %ld.%03ld secs.\n",
	  status, timer.expend.tv_sec, timer.expend.tv_usec / 1000);
#endif	/* DEBUG */

  netfd_finish();

  return status;
}
